@@ -1,12 +1,16 @@
 import os
 from pathlib import Path
 from string import Template
+from typing import Type, List, TYPE_CHECKING
 
 import ipywidgets.widgets as widgets
 import ujson
 from IPython.core.display import Javascript
 from ipycallback import SlowTunnelWidget
-from pandas import DataFrame
+
+if TYPE_CHECKING:
+    # do this because pandas is optional
+    from pandas import DataFrame
 
 from labext.helpers import read_file
 from labext.widget import WidgetWrapper
@@ -14,7 +18,7 @@ import labext.modules as M
 
 
 class DataTable(WidgetWrapper):
-    def __init__(self, df: DataFrame, table_class: str='display', **kwargs):
+    def __init__(self, df: 'DataFrame', table_class: str='display', **kwargs):
         """Show the
 
         Parameters
@@ -82,6 +86,10 @@ class DataTable(WidgetWrapper):
     def widget(self):
         return self.el
 
+    @staticmethod
+    def required_modules() -> List[Type[M.Module]]:
+        return [M.DataTable, M.JQuery, M.LabExt]
+
     def get_auxiliary_components(self, *args) -> list:
         return self.el_auxiliaries
 
