@@ -26,7 +26,6 @@ class Selection(WidgetWrapper):
         self.el_root_id = f"selection_{self.el_root.model_id}"
         self.el_root.add_class(self.el_root_id)
 
-        self.version = -1
         # default value is '' (no choice) rather than None to be consistent with the selectize library!
         self.value: str = default_value
 
@@ -42,7 +41,7 @@ class Selection(WidgetWrapper):
 
         self.el_search_tunnel = SlowTunnelWidget()
         self.el_value_tunnel = SlowTunnelWidget()
-        self.el_value_tunnel.on_receive(self._handle_change)
+        self.el_value_tunnel.on_receive(self.on_receive_updates)
 
         if self.search_fn is not None:
             assert len(
@@ -51,6 +50,9 @@ class Selection(WidgetWrapper):
             self.records = self.search_fn("")
             self.el_search_tunnel.on_receive(self._handle_search)
 
+        self.pending_ops = []
+        self.is_frontend_initialized = False
+
     @property
     def widget(self):
         return self.el_root
@@ -63,24 +65,68 @@ class Selection(WidgetWrapper):
         results = self.search_fn(query)
         self.el_search_tunnel.push(version, ujson.dumps(results))
 
-    def _handle_change(self, version: int, value: str):
-        if version > self.version:
-            self.version = version
-            if value.startswith("["):
-                value = ujson.loads(value)
+    def on_receive_updates(self, version: int, msg: str):
+        msg = ujson.loads(msg)
+        if msg['type'] == 'initialized':
+            self.is_frontend_initialized = True
+            if len(self.pending_ops) > 0:
+                for op, args in self.pending_ops:
+                    if op == 'set_value':
+                        self.set_value(*args)
+                    elif op == 'replace_options':
+                        self.replace_options(*args)
+                    elif op == 'add_options':
+                        self.add_options(*args)
+                self.pending_ops = []
+            return
 
-            self.value = value
-            self.on_change_callback(value)
+        if msg['type'] == 'set_value':
+            self.value = msg['value']
+            self.on_change_callback(self.value)
+            return
 
     def get_value(self):
         return self.value
 
     def set_value(self, value):
-        self.version += 1
+        if not self.is_frontend_initialized:
+            self.pending_ops.append(["set_value", (value,)])
+            return
+
         self.value = value
-        if isinstance(self.value, list):
-            value = ujson.dumps(self.value)
-        self.el_value_tunnel.push(self.version, value)
+        msg = {
+            "type": "set_value",
+            "value": self.value
+        }
+        self.el_value_tunnel.send_msg(ujson.dumps(msg))
+
+    def replace_options(self, records: List[dict]):
+        if not self.is_frontend_initialized:
+            self.pending_ops.append(["replace_options", (records,)])
+            return
+
+        self.records = records
+        msg = {
+            "type": "replace_options",
+            "records": records
+        }
+        self.el_value_tunnel.send_msg(ujson.dumps(msg))
+
+    def add_options(self, records: List[dict]):
+        if not self.is_frontend_initialized:
+            self.pending_ops.append(["add_options", (records,)])
+            return
+
+        ids = {r[self.value_field] for r in self.records}
+        for r in records:
+            if r[self.value_field] not in ids:
+                self.records.append(r)
+
+        msg = {
+            "type": "add_options",
+            "records": records
+        }
+        self.el_value_tunnel.send_msg(ujson.dumps(msg))
 
     def get_auxiliary_components(self):
         selectize_options = """
@@ -92,11 +138,10 @@ class Selection(WidgetWrapper):
             maxItems: $maxItems,
             items: $items,
             onChange: function (value) {
-                if (Array.isArray(value)) {
-                    window.IPyCallback.get("$valueTunnel").send_msg(JSON.stringify(value));
-                } else {
-                    window.IPyCallback.get("$valueTunnel").send_msg(value);
-                }
+                window.IPyCallback.get("$valueTunnel").send_msg(JSON.stringify({
+                    type: "set_value",
+                    value: value
+                }));
             },
             render: {
                 item: function (item, escape) {
@@ -148,19 +193,33 @@ require(["$JQueryId", "$SelectizeId"], function (jquery, _) {
         });
 
         var selectize = jquery('select', $$el)[0].selectize;
-        window.IPyCallback.get("$valueTunnel").on_receive((version, value) => {
-            if (value == "") {
-                selectize.clear(true);
-            } else if (value[0] == "[") {
-                var vals = JSON.parse(value);
-                selectize.clear(true);
-                for (let v of vals) {
-                    selectize.addItem(v, true);
+        window.IPyCallback.get("$valueTunnel").on_receive((version, payload) => {
+            let msg = JSON.parse(payload);
+            if (msg.type == 'set_value') {
+                if (msg.value == '') {
+                    selectize.clear(true);  
+                } else if (Array.isArray(msg.value)) {
+                    selectize.clear(true);
+                    for (let v of msg.value) {
+                        selectize.addItem(v, true);
+                    }
+                } else {
+                    selectize.addItem(msg.value, true);
                 }
-            } else {
-                selectize.addItem(value, true);
+            } else if (msg.type == 'replace_options') {
+                selectize.clearOptions();
+                for (let record of msg.records) {
+                    selectize.addOption(record);
+                }
+                selectize.refreshOptions(false);
+            } else if (msg.type == 'add_options') {
+                for (let record of msg.records) {
+                    selectize.addOption(record);
+                }
+                selectize.refreshOptions(false);
             }
         });
+        window.IPyCallback.get("$valueTunnel").send_msg(JSON.stringify({"type": "initialized"}));
         return true;
     }, 100);
 });
